version: str = '0.9.2'

name: str = 'alteza'

# pylint: disable=consider-using-f-string
repo_url: str = 'https://github.com/arjun-menon/%s' % name
download_url: str = '%s/archive/v%s.tar.gz' % (repo_url, version)

author: str = 'Arjun G. Menon'
author_email: str = 'contact@arjungmenon.com'
